package eu.doppel_helix.dev.blaesing.testbindingio.testbindingio;


import com.sun.jna.Function;
import com.sun.jna.Native;
import com.sun.jna.NativeLibrary;
import com.sun.jna.Pointer;

public class TestIndirect {

    private static final NativeLibrary libc = NativeLibrary.getInstance("c");

    private static Function fopen64;
    private static Function fopen;
    private static Function fclose;
    private static Function fread;
    private static Function fwrite;
    private static Function fseek;
    private static Function mmap;
    private static Function mmap64;
    private static Function munmap;
    private static Function open;
    private static Function open64;
    private static Function close;

    /**
     * Seek from beginning of file.
     */
    public static final int SEEK_SET = 0;
    /**
     * Seek from current position.
     */
    public static final int SEEK_CUR = 1;
    /**
     * Seek from end of file.
     */
    public static final int SEEK_END = 2;

    /**
     * Page can be read.
     */
    public static final int PROT_READ = 0x1;
    /**
     * Page can be written.
     */
    public static final int PROT_WRITE = 0x2;
    /**
     * Page can be executed.
     */
    public static final int PROT_EXEC = 0x4;
    /**
     * Page can not be accessed.
     */
    public static final int PROT_NONE = 0x0;
    /**
     * Extend change to start of growsdown vma (mprotect only).
     */
    public static final int PROT_GROWSDOWN = 0x01000000;
    /**
     * Extend change to start of growsup vma (mprotect only).
     */
    public static final int PROT_GROWSUP = 0x02000000;
    public static final int O_RDONLY = 00000000;
    public static final int O_WRONLY = 00000001;
    public static final int O_RDWR = 00000002;

    /**
     * Share changes.
     */
    public static final int MAP_SHARED = 0x01;
    /*
     * Changes are private.
     */
    public static final int MAP_PRIVATE = 0x02;


    static {
        try {
            fopen64 = libc.getFunction("fopen64", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            fopen = libc.getFunction("fopen", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            fread = libc.getFunction("fread", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            fclose = libc.getFunction("fclose", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            fwrite = libc.getFunction("fwrite", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            fseek = libc.getFunction("fseek", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            mmap = libc.getFunction("mmap", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            mmap64 = libc.getFunction("mmap64", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            munmap = libc.getFunction("munmap", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            open = libc.getFunction("open", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            open64 = libc.getFunction("open64", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
        try {
            close = libc.getFunction("close", Function.THROW_LAST_ERROR);
        } catch (UnsatisfiedLinkError ex) {
        }
    }

    private static Pointer fopen(String path, String mode) {
        if(fopen64 != null) {
            return fopen64.invokePointer(new Object[] {path, mode});
        } else {
            return fopen.invokePointer(new Object[] {path, mode});
        }
    }

    private static int open(String path, int mode) {
        if(open64 != null) {
            return open64.invokeInt(new Object[] {path, mode});
        } else {
            return open.invokeInt(new Object[] {path, mode});
        }
    }

    private static int close(int fd) {
        return close.invokeInt(new Object[]{fd});
    }

    private static int fclose(Pointer fid) {
        return fclose.invokeInt(new Object[] {fid});
    }

    private static long fread(Pointer output, long size, long memb, Pointer fid) {
        if(Native.SIZE_T_SIZE == 4) {
            require32Bit(memb, "memb");
            require32Bit(size, "size");
            return fread.invokeInt(new Object[] {output, (int) size, (int) memb, fid});
        } else {
            return fread.invokeLong(new Object[] {output, size, memb, fid});
        }
    }

    private static long fwrite(Pointer output, long size, long memb, Pointer fid) {
        if(Native.SIZE_T_SIZE == 4) {
            require32Bit(memb, "memb");
            require32Bit(size, "size");
            return fwrite.invokeInt(new Object[] {output, (int) size, (int) memb, fid});
        } else {
            return fwrite.invokeLong(new Object[] {output, size, memb, fid});
        }
    }

    private static int fseek(Pointer fid, long offset, int whence) {
        if(Native.LONG_SIZE == 4) {
            require32Bit(offset, "offset");
            return fseek.invokeInt(new Object[] {fid, (int) offset, whence});
        } else {
            return fseek.invokeInt(new Object[] {fid, offset, whence});
        }
    }

    private static Pointer mmap(Pointer addr, long length, int prot, int flags, int fid, long offset) {
        Object[] params = new Object[6];
        params[0] = addr;
        if(Native.SIZE_T_SIZE == 4) {
            require32Bit(length, "length");
            params[1] = (int) length;
        } else {
            params[1] = length;
        }
        params[2] = prot;
        params[3] = flags;
        params[4] = fid;
        if(mmap64 != null) {
            params[5] = offset;
            return mmap64.invokePointer(params);
        } else {
            if(Native.LONG_SIZE == 4) {
                // Observations shows, that on linux, solaris, aix, mac OS and windows (apart from cygwin32)
                // sizeof(off_t) == sizeof(long)
                params[5] = (int) offset;
            } else {
                params[5] = offset;
            }
            return mmap.invokePointer(params);
        }
    }

    private static int munmap(Pointer addr, long length) {
        if (Native.SIZE_T_SIZE == 4) {
            require32Bit(length, "length");
            return munmap.invokeInt(new Object[]{addr, (int) length});
        } else {
            return munmap.invokeInt(new Object[]{addr, length});
        }
    }

    public static void main(String[] argv) {

//        Pointer fid = fopen("/home/matthias/test", "r+");
//        Memory buffer = new Memory(1024);
//        buffer.clear();
//        long read = fread(buffer, 1, buffer.size(), fid);
//        System.out.println(read);
//        System.out.println(new String(buffer.getByteArray(0, (int) read)));
//        fseek(fid, 0, SEEK_SET);
//        buffer.setString(0, "Demo String");
//        fwrite(buffer, 1, 11, fid);
//        fclose(fid);

        int fd = open("/home/matthias/test", O_RDONLY);
        Pointer mapped = mmap(null, 1024, PROT_READ, MAP_PRIVATE, fd, 0);
        close(fd);

        System.out.println(new String(mapped.getByteArray(0, 10)));

        munmap(mapped, 1024);

    }

    private static void require32Bit(long val, String value) {
        if(val > Integer.MAX_VALUE) {
            throw new IllegalArgumentException(value + " exceeds 32bit");
        }
    }
}
